import unittest
from natia.kb.docdb import MongoDBClient

class DocumentDatabaseTester(unittest.TestCase):
    
    def test_():
        
        ...
        
if __name__ == '__main__':
    
    unittest.main()